//	spearsynth4~ -- use sine waves to do a crude SPEAR resynth, this with controls
//	first inlet -- 1 to start/0 to stop; second inlet -- frame number (0-999)
//			Brad Garton, 12/2005


#include "ext.h"
#include "z_dsp.h"
#include "string.h"
#include "ext_strings.h"
#include "edit.h"
#include "ext_wind.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define MAX_OUTPUTS 2	// also arbitrary
#define NUM_INLETS 4 // for my inlets -- one for stap/start; 1 for frameno; 2 for R; 3 for hz spread
// we are total memory hogs with the #defines below -- this could be handled much much better
#define MAX_PARTIALS 500 // this is how many total partials we can read in at any one time.
#define MAX_FRAMES 1000 // allow for 10 seconds

// oh I am a bad brad programmer here!  These *need* to be allocatedm and inside the t_spearsynth4 structure
// right now only 1 spearsynth4~ will operate
float si[MAX_FRAMES][MAX_PARTIALS]; // store the sampling increments used for the oscillator frequency
float phase[MAX_FRAMES][MAX_PARTIALS]; // store where we currently are located in the sinetable[]
float amp[MAX_FRAMES][MAX_PARTIALS]; // amplitudes for the partials
float pitchdev[MAX_PARTIALS]; // this is how much we will multiply each partial frequency

void *spearsynth4_class;

// declare the sine wave table out here so that all instances of the spearsynth4~ object
// can reference it; that way we don't create it anew for each object.
#define MAXSINETABLE 44100 // make it easy for us -- 1 cycle/second; si then equals freq
float sinetable[MAXSINETABLE];

// in this structure are variables specific to each instance of spearsynth4~
typedef struct _spearsynth4
{
	// header
	t_pxobject x_obj;

	long num_outputs; 	// number of outputs (mono or stereo)
	long num_inlets;	// we'll set this statically for our controlling inlets

	int goflag; // when "1", synthesis starts.  Set by a bang.
	int numframes; // total number of frames to be used
	int curframe; // current frame data being used
	int data_read; // set to "1" when data has been read

	// these are for the CHAOS!!!! WHOOOOOOOOO!!!!
	float R;
	float spread;
	float xval;
} t_spearsynth4;


/****PROTOTYPES****/

//	max/msp setup funcs; this probably won't change, unless you decide to change the number of
//	args that the user can input, in which case spearsynth4_new will have to change
void *spearsynth4_new(long num_outputs);	// this allows us to have one optional params for the object
void spearsynth4_dsp(t_spearsynth4 *x, t_signal **sp, short *count); 
t_int *spearsynth4_perform(t_int *w);
void spearsynth4_assist(t_spearsynth4 *x, void *b, long m, long a, char *s);

// the following are functions we define below for handling floats/ints/bangs
// at the inputs.  We can delete these if we don't use them
void spearsynth4_float(t_spearsynth4 *x, double f);
void spearsynth4_int(t_spearsynth4 *x, int i);
void spearsynth4_bang(t_spearsynth4 *x);

// the functions for reading in the SPEAR data file -- doread() is so we can defer()
void spearsynth4_read(t_spearsynth4 *x, Symbol *s, short argc, Atom *argv);
void spearsynth4_doread(t_spearsynth4 *x, Symbol *s, short argc, t_atom *argv);

/****FUNCTIONS****/

//primary MSP funcs -- this is called the first time you instantiate an object of this type
void main(void)
{
	int i;
	float radians;
	
	// the A_DEFLONG argument sets mono or stereo output (default will be mono)
	setup((struct messlist **)&spearsynth4_class, (method)spearsynth4_new, (method)dsp_free, (short)sizeof(t_spearsynth4), 0L, A_DEFLONG, 0);

	// max/msp standard messages; don't change these  
	addmess((method)spearsynth4_dsp, "dsp", A_CANT, 0);
	addmess((method)spearsynth4_assist,"assist", A_CANT,0);

	// spearsynth4_float will be called for any floats coming in any inlet
	addfloat((method)spearsynth4_float);

	// spearsynth4_stopstart will be called for ints coming in inlet 0
	addint((method)spearsynth4_int);

	// spearsynth4_bang will be called for any bangs coming in any inlet
	addbang((method)spearsynth4_bang);

	// this is for the [read ...] message
	addmess((method)spearsynth4_read, "read", A_GIMME, 0);

	// msp initialization (gotta have this)
	dsp_initclass();

	// set up the sine wavetable
	for (i = 0; i < MAXSINETABLE; i++) {
		radians = (float)i/(float)MAXSINETABLE * TWOPI;
		sinetable[i] = sin(radians);
	}

	post("spearsynth4~ external loaded!");
}


// this gets called when the object is created and every time the user types in new args
void *spearsynth4_new(long num_outputs)
{
	int i,j;

	// leave this; creates our object
	t_spearsynth4 *x = (t_spearsynth4 *)newobject(spearsynth4_class);

	// zero out the struct, to be careful (takk to jkclayton)
	if (x) { 
		for (i = sizeof(t_pxobject); i < sizeof(t_spearsynth4); i++)  
			((char *)x)[i] = 0;
	}

	// this is where we access the values (if any) of the optional params for the object	
	// constrain number of outputs
	if (num_outputs < 1) num_outputs = 1;
	if (num_outputs > MAX_OUTPUTS) num_outputs = 1;

	// and we set the internal t_spearsynth4 structure variables
	x->num_outputs = num_outputs;
	x->num_inlets = NUM_INLETS; // left for stop/start; right for frameno
	x->curframe = 0; // set as default
	x->goflag = 0; // "1" starts synthesis output (set from inlet 1)
	x->data_read = 0; // set to "1" after reading data file
	x->xval = 0.5; // for the chaos

	// now setup audio inputs and outputs
	// inputs
	dsp_setup((t_pxobject *)x, x->num_inlets);
   
	// signal outputs
	for (i = 0; i < x->num_outputs; i++) {
		outlet_new((t_object *)x, "signal");
	}

	//occasionally this line is necessary if you are doing weird asynchronous things with the in/out vectors
	//x->x_obj.z_misc = Z_NO_INPLACE;

	return (x);
}


// this gets called everytime audio is started; even when audio is running, if the user changes anything
// (like deletes a patch cord), audio will be turned off and then on again, calling this func.
// this adds the "perform" method to the DSP chain, and also tells us
// where the audio vectors are and how big they are
void spearsynth4_dsp(t_spearsynth4 *x, t_signal **sp, short *count)
{
	void *dsp_add_args[NUM_INLETS + MAX_OUTPUTS + 2];
	int i;

	// construct the array of vectors and stuff for max/msp to knoww
	dsp_add_args[0] = x; // the object itself
	for (i = 0; i < (x->num_inlets + x->num_outputs); i++) { // pointers to the input and output vectors
		dsp_add_args[i+1] = sp[i]->s_vec;
	}
	dsp_add_args[x->num_inlets+ x->num_outputs + 1] = (void *)sp[0]->s_n; // pointer to the vector size
	dsp_addv(spearsynth4_perform, (x->num_inlets + x->num_outputs + 2), dsp_add_args); // add them to the signal chain

	x->goflag = 0; // "1" starts synthesis output, from a bang
}


// this is where the action is
// we get vectors of samples (n samples per vector), process them and send them out
// the "w" array contains a lot of stuff (note no input pointers in this one!)
//	w[0] -- pointer to the spearsynth4_perform function being called
//	w[1] -- pointer to the t_spearsynth4 object structure data
//	w[2] ... w[numinlets+1] -- pointers to inlets (unused for signal here)
//	w[numinlets+1] ... w[numoutputs+ whatever] -- pointers to arrays to be filled with samples (output arrays)
//	w[numoutputs+2] -- number of samples/vector (i.e. the number of samples in each input or output array)
t_int *spearsynth4_perform(t_int *w)
{
	t_spearsynth4 *x = (t_spearsynth4 *)(w[1]);

	float *out[MAX_OUTPUTS];	// pointers to the output vectors
	// note the the above are pointers -- each output is represented as an array
	// of samples (i.e. non-interleaved).  These are set to point to the w[] arrays.

	long n = w[x->num_inlets + x->num_outputs + 2];	// number of samples per vector

	// local vars
	int i;

	// check to see if we should skip this routine if the patcher is "muted"
	// all audio objects should have this, even though the 'goto' is supremely elegant.
	if (x->x_obj.z_disabled) goto out;

	// assign the output vector pointers
	for (i = 0; i < x->num_outputs; i++) {
		out[i] = (float *)(w[x->num_inlets+i+2]);
	}
	// after this loop above, we can write samples using *out[oulet]+sampval to
	// get and set sample values for each outlet vector.


/* ------- this is the sample computing loop for each vector ------- */
	while (n--) { // n is the number of samples for the vector

		*out[0] = 0.0;
		if (x->goflag == 1) {
			for (i = 0; i < MAX_PARTIALS; i++) {
				if (amp[x->curframe][i] > 0.0) {
					*out[0] += (sinetable[(int)phase[0][i]] * amp[x->curframe][i]);
					phase[0][i] += (si[x->curframe][i] + pitchdev[i]);
					if (phase[0][i] > (float)MAXSINETABLE)
						phase[0][i] = phase[0][i] - (float)MAXSINETABLE;
				}
			}
		}
		
		if (x->num_outputs == 2) *out[1] = *out[0];
		*out[0]++;
		*out[1]++;
	}
/* ------- end of the sample computing loop ------- */


	// return a pointer to the next object in the signal chain.
out:
	return w + x->num_inlets + x->num_outputs + 3;
}


// tells the user about the inputs/outputs when mousing over them
// "m" is inlet (1) or outlet (2); "a" is which one the mouse is over
void spearsynth4_assist(t_spearsynth4 *x, void *b, long m, long a, char *s)
{
	int i;
	
	//could use switch/case inside for loops, to give more informative assist info....
	if (m == 1) {
			if (a == 0) sprintf(s, "0-stop/1-start");
			if (a == 1) sprintf(s, "frame #");
	}
	if (m == 2) {
		for (i = 0; i <= x->num_outputs; i++) // "<=" because we have one additional outlet for bangs/vals
			if (a == i) sprintf(s, "signal %d", i);  // ditto about message that gets shown
	}
}


// this gets called whenever a float is received at *any* input
void spearsynth4_float(t_spearsynth4 *x, double f)
{
	int i;

	switch (x->x_obj.z_in) {
		case 2:
			if (f < 0.0) f = 0.0;
			if (f > 3.99) f = 3.99;
			x->R = f;
			break;
		case 3:
			if (f < 0.0) f = 0.0;
			x->spread = f;	
			break;
		default:
			return;
	}

	x->xval = 0.5;
	for (i = 0; i < 1000; i++) {
		x->xval = (x->R*x->xval)*(1.0 - x->xval);
	}

	for (i = 0; i < MAX_PARTIALS; i++) {
		x->xval = (x->R*x->xval)*(1.0 - x->xval);
		pitchdev[i] = x->xval * x->spread;
	}
}


// this gets called whenever an int is received at *any* input
void spearsynth4_int(t_spearsynth4 *x, int k)
{
	switch (x->x_obj.z_in) {
		case 0:
			if (k == 0) x->goflag = 0;
			else {
				if (x->data_read == 0) {
					post("spearsynth4~: you must read in a SPEAR data file before running");
					return;
				} else {
					x->goflag = 1;
				}
			}
			break;
		case 1:
			if (k < MAX_FRAMES) x->curframe = k;
			else x->curframe = MAX_FRAMES-1;
	}	
}


// we're going to use the bang to reset to the beginning of the SPEAR data and trigger synthesis
// this gets called whenever a bang is received at *any* input
void spearsynth4_bang(t_spearsynth4 *x)
{
	int i;

	for (i = 0; i < 1000; i++) {
		x->xval = (x->R*x->xval)*(1.0 - x->xval);
	}

	for (i = 0; i < MAX_PARTIALS; i++) {
		x->xval = (x->R*x->xval)*(1.0 - x->xval);
		pitchdev[i] = x->xval * x->spread;
	}
}


// the [read] message reads in a SPEAR resampleframes text file, and sets a pointer to the
// memory for use in the spearsynth4_perform() function.  The read gets deferred because it
// is a slow operation
void spearsynth4_read(t_spearsynth4 *x, Symbol *s, short argc, Atom *argv)  
{
	defer(x, (method)spearsynth4_doread, s, argc, argv); // always defer this message  
}

// this opens and reads in a file
void spearsynth4_doread(t_spearsynth4 *x, Symbol *s, short argc, t_atom *argv) 
{
	char filename[256];
	int nameflag; // we use this to determine if we should open a dialog
	short err, i, j;
	short path;
	long type = 'TEXT'; 
	long size;
	long outtype, count;
	t_filehandle fh; // this to reference the SPEAR data file
	t_handle speardatahandle; // use this to reference the SPEAR data we read in

	// these are for the data-reading/parsing ugliness
	char *bufptr;
	char parsebuf[1000];
	int linepartials, pno;
	
	nameflag = 0;	
	for (i = 0; i < argc; i++) {
		switch (argv[i].a_type) {
			case A_LONG:
				post("spearsynth4 read should be a single message or contain a filename");
				return;
			case A_FLOAT:
				post("spearsynth4 read should be a single message or contain a filename");
				return;
			case A_SYM:
				strcpy(filename, argv[i].a_w.w_sym->s_name);
				nameflag = 1;
		}
	}
	
	if (nameflag == 0) {
		if (open_dialog(filename,  &path, &outtype, 0L, 0)) // allow all types of files
			return; //user cancelled
	} else {
		if (locatefile_extended(filename, &path, &outtype, &type, 1)) {
			error("spearsynth4~: error opening file: can't find file"); 
			return; //not found
		}
	}
	
	//we should have a valid filename at this point
	err = path_opensysfile(filename, path, &fh, READ_PERM);  
	if (err) {       
		error("error %d opening file", err); 
		return;
	}
	
	sysfile_geteof(fh, &size); // get the size of the file
	if (!(speardatahandle = sysmem_newhandle(size))) { // allocate the memory
		error("spearsynth4~: %s too big to read", filename);
		return;
	} else {
		err = sysfile_readtextfile(fh, speardatahandle, size, TEXT_LB_NATIVE); // read that file!
		if (err) {
			error("trouble %d reading SPEAR data file", err);
			return;
		}
	}

	// now we're going to go through the data we just read and grab the info we need
	// and store them in our arrays.  jeez, this is truly insane...
	bufptr = (char *)*speardatahandle; // bufptr initially set to the beginning of the data
	
	for (i = 0; i < 8; i++) {
		sscanf(bufptr, "%s ", parsebuf);
		bufptr += (strlen(parsebuf)+1); // get the 'white space'
	}

	// the next item should be the frame-count number
	sscanf(bufptr, "%s", parsebuf);
	bufptr += (strlen(parsebuf)+1);
	sscanf(parsebuf, "%d", &(x->numframes));
	if (x->numframes > MAX_FRAMES) {
		post("spearsynth4~ can only do %d frames of data, limiting to that amount", MAX_FRAMES);
		x->numframes = MAX_FRAMES;
	}

	// go one more...
	sscanf(bufptr, "%s", parsebuf);
	bufptr += (strlen(parsebuf)+1);

	// now we should be positioned at the beginning of the frame data

	// initialize amp and phase for this data -- we use amp to determine synthesis
	for (i = 0; i < MAX_FRAMES; i++) {
		for (j = 0; j < MAX_PARTIALS; j++) {
			amp[i][j] = 0.0;
		}
	}

	// now get the amp and freq (si) values for the appropriate partials
	for (i = 0; i < x->numframes; i++) {
		// read the frame time-val (unused) and the number of partials for this line
		sscanf(bufptr, "%s", parsebuf);
		bufptr += (strlen(parsebuf)+1);
		sscanf(bufptr, "%s", parsebuf);
		bufptr += (strlen(parsebuf)+1);
		sscanf(parsebuf, "%d", &linepartials);

		for (j = 0; j < linepartials; j++) {
			// get the partial #
			sscanf(bufptr, "%s", parsebuf);
			bufptr += (strlen(parsebuf)+1);
			sscanf(parsebuf, "%d", &pno);
			// and the freq (si)
			sscanf(bufptr, "%s", parsebuf);
			bufptr += (strlen(parsebuf)+1);
			if (pno < MAX_PARTIALS) {
				sscanf(parsebuf, "%f", si[i]+pno);
			} else {
				post("spearsynth4~ can only do %d partials, limiting to that amount", MAX_PARTIALS);
			}
			// and the amp
			sscanf(bufptr, "%s", parsebuf);
			bufptr += (strlen(parsebuf)+1);
			if (pno < MAX_PARTIALS) {
				sscanf(parsebuf, "%f", amp[i]+pno);
			}
			// and set the phase to 0 for this one
			if (pno < MAX_PARTIALS) {
				phase[0][pno] = 0.0;
			}
		}
	}

	// free up the resources we used for this read operation
	sysmem_freehandle(speardatahandle);
	sysfile_close(fh);

	x->data_read = 1; // all set!

	return;
}
